from unittest import TestCase

from doxyqml.lexer import Lexer, Token, IMPORT, PRAGMA, STRING, COMMENT, KEYWORD, ELEMENT, \
        BLOCK_START, COMPONENT


class LexerTestCase(TestCase):
    def test_import(self):
        src = "import foo\n import bar"
        lexer = Lexer(src)
        lexer.tokenize()
        self.assertEqual(lexer.tokens[0], Token(IMPORT, "import foo", 0))
        self.assertEqual(lexer.tokens[1], Token(IMPORT, "import bar", 12))

    def test_pragma(self):
        src = "pragma foo\n pragma bar"
        lexer = Lexer(src)
        lexer.tokenize()
        self.assertEqual(lexer.tokens[0], Token(PRAGMA, "pragma foo", 0))
        self.assertEqual(lexer.tokens[1], Token(PRAGMA, "pragma bar", 12))

    def test_string(self):
        src = r'"hello" "world!" "new\nline" "qu\"ote"'
        lexer = Lexer(src)
        lexer.tokenize()
        self.assertEqual(lexer.tokens[0], Token(STRING, '"hello"', 0))
        self.assertEqual(lexer.tokens[1], Token(STRING, '"world!"', 8))
        self.assertEqual(lexer.tokens[2], Token(STRING, r'"new\nline"', 17))
        self.assertEqual(lexer.tokens[3], Token(STRING, r'"qu\"ote"', 29))

    def test_single_line_comment(self):
        src = "// hello\nimport bob"
        lexer = Lexer(src)
        lexer.tokenize()
        self.assertEqual(lexer.tokens[0], Token(COMMENT, '// hello', 0))
        self.assertEqual(lexer.tokens[1], Token(IMPORT, 'import bob', 9))

    def test_multi_line_comment(self):
        src = "/* hello\nworld *//* good bye\nworld */"
        lexer = Lexer(src)
        lexer.tokenize()
        self.assertEqual(lexer.tokens[0], Token(COMMENT, '/* hello\nworld */', 0))
        self.assertEqual(lexer.tokens[1], Token(COMMENT, '/* good bye\nworld */', 17))

    def test_property_named_property(self):
        src = "Item { property var property }"
        lexer = Lexer(src)
        lexer.tokenize()
        self.assertEqual(lexer.tokens[0], Token(COMPONENT, 'Item', 0))
        self.assertEqual(lexer.tokens[1], Token(BLOCK_START, '{', 5))
        self.assertEqual(lexer.tokens[2], Token(KEYWORD, 'property', 7))
        self.assertEqual(lexer.tokens[3], Token(ELEMENT, 'var', 16))
        self.assertEqual(lexer.tokens[4], Token(ELEMENT, 'property', 20))



import re
from unittest import TestCase

from doxyqml.qmlclass import QmlFunction, QmlArgument, QmlProperty


class QmlFunctionTestCase(TestCase):
    def test_post_process_doc_at(self):
        self._test_post_process_doc("""
        /**
         * Create a user
         *
         * @param type:string firstname The user firstname
         * @param type:string lastname The user lastname
         * @param type:int age The user age
         * @param misc A parameter with no type
         * @return type:User A new user
         */
         """)

    def test_post_process_doc_backslash(self):
        self._test_post_process_doc(r"""
        /**
         * Create a user
         *
         * \param type:string firstname The user firstname
         * \param type:string lastname The user lastname
         * \param type:int age The user age
         * \param misc A parameter with no type
         * \return type:User A new user
         */
         """)

    def _test_post_process_doc(self, doc):
        fcn = QmlFunction()
        fcn.args = [
            QmlArgument("firstname"),
            QmlArgument("lastname"),
            QmlArgument("age"),
            QmlArgument("misc"),
            ]
        fcn.doc = doc
        fcn.post_process_doc()

        self.assertEqual(fcn.args[0].type, "string")
        self.assertEqual(fcn.args[1].type, "string")
        self.assertEqual(fcn.args[2].type, "int")
        self.assertEqual(fcn.args[3].type, "")
        self.assertEqual(fcn.type, "User")

        # "[@\]param type:..." are turned into "@param"
        expected_doc = re.sub(r"[@\\]param type:\w+", r"@param", doc)
        # "[@\]return type:..." are turned into "[@\]return"
        expected_doc = re.sub(r"([@\\])return type:\w+", r"\1return", expected_doc)
        self.assertMultiLineEqual(fcn.doc, expected_doc)


class QmlPropertyTestCase(TestCase):
    def test_property_type(self):
        prop = QmlProperty()
        prop.doc = "/// type:User The current user"
        prop.type = "alias"

        prop.post_process_doc()

        self.assertEqual(prop.type, "User")
        self.assertEqual(prop.doc, "/// The current user")

    def test_no_property_type(self):
        prop = QmlProperty()
        prop.doc = "/// The user age"
        prop.type = "int"

        prop.post_process_doc()

        self.assertEqual(prop.type, "int")
        self.assertEqual(prop.doc, "/// The user age")

    def test_default_property(self):
        prop = QmlProperty()
        prop.doc = "/// Children"
        prop.type = "list<Item>"
        prop.is_default = True

        self.assertEqual(str(prop),
                         "/// Children\n" + QmlProperty.DEFAULT_PROPERTY_COMMENT + "\nQ_PROPERTY(list<Item>  READ dummyGetter__ignore)")



import os

def pytest_generate_tests(metafunc):
    script_dir = os.path.dirname(__file__) or "."
    os.chdir(script_dir)
    test_list = [x for x in os.listdir("./functional/") if os.path.isdir(x)]
    metafunc.parametrize("directory", test_list)



import logging
import re
import typing

TYPE_RX = r"(?P<prefix>\s+type:)(?P<type>[\w.<>|]+)"

BASE_NAME_DICT = {
    "QtObject": "QtQml.QtObject",
    "Item": "QtQuick.Item",
}


def post_process_type(rx, text, type):
    match = rx.search(text)
    if match:
        type = match.group("type")
        text = text[:match.start("prefix")] + text[match.end("type"):]
    return text, type


class QmlBaseComponent():
    def __init__(self, name):
        self.name = name
        self.base_name = ""
        self.elements = []

        lst = name.split(".")
        self.class_name = lst[-1]
        self.namespaces = lst[:-1]

    def get_attributes(self):
        return [x for x in self.elements if isinstance(x, QmlAttribute)]

    def get_properties(self):
        return [x for x in self.elements if isinstance(x, QmlProperty)]

    def get_functions(self):
        return [x for x in self.elements if isinstance(x, QmlFunction)]

    def get_signals(self):
        return [x for x in self.elements if isinstance(x, QmlSignal)]

    def add_element(self, element):
        self.elements.append(element)

    def __str__(self):
        lst = []
        self._export_content(lst)
        return "\n".join(lst)

    def _export_elements(self, input_list, lst, filter=None):
        for element in input_list:
            if filter and not filter(element):
                continue
            doc = str(element)
            if doc:
                lst.append(doc)

    def _start_class(self, lst):
        class_decl = "class " + self.class_name
        if self.base_name:
            for alias, replacement in self.alias.items():
                self.base_name = re.sub(alias, replacement, self.base_name)
            self.base_name = BASE_NAME_DICT.get(self.base_name, self.base_name)

            class_decl += " : public " + self.base_name

        class_decl += " {"
        lst.append(class_decl)
        lst.append("public:")

    def _end_class(self, lst):
        lst.append("};")


class QmlClass(QmlBaseComponent):
    SINGLETON_COMMENT = "/** @remark This component is a singleton */"
    VERSION_COMMENT = "/** @since %s */"

    def __init__(self, name, version=None):
        QmlBaseComponent.__init__(self, name)
        self.header_comments = []
        self.footer_comments = []
        self.imports = []
        self.alias = {}
        if version:
            self.header_comments.append(QmlClass.VERSION_COMMENT % version)

    def add_pragma(self, decl):
        args = decl.split(' ', 2)[1].strip()

        if args.lower() == "singleton":
            self.header_comments.append(QmlClass.SINGLETON_COMMENT)

    def add_import(self, decl):
        modules = decl.split()
        module = modules[1]
        if module[0] == '"':
            # Ignore directory or javascript imports for now
            return
        if len(modules) > 3:
            self.alias[modules[4]] = modules[1]
        self.imports.append(module)

    def add_header_comment(self, obj):
        self.header_comments.append(obj)

    def add_footer_comment(self, obj):
        self.footer_comments.append(obj)

    def _export_header(self, lst):
        for module in self.imports:
            lst.append("using namespace %s;" % module.replace('.', '::'))
        if self.namespaces:
            lst.append("namespace %s {" % '::'.join(self.namespaces))

        lst.extend([str(x) for x in self.header_comments])

    def _export_footer(self, lst):
        lst.extend([str(x) for x in self.footer_comments])

        if self.namespaces:
            lst.append("}")

    def _export_content(self, lst):
        self._export_header(lst)

        # Public members.
        self._start_class(lst)

        public_members = []
        private_members = []

        # Sort elements before exporting to reduce the number of times element list must
        # be iterated through.
        last_element_was_public = True
        for element in self.elements:
            if str(element) == "" or isinstance(element, str) :
                if last_element_was_public:
                    public_members.append(element)
                else:
                    private_members.append(element)
            elif element.is_public_element():
                public_members.append(element)
                last_element_was_public = True
            else:
                private_members.append(element)
                last_element_was_public = False

        self._export_elements(public_members, lst)

        if len(private_members) > 0:
            lst.append("private:")
            self._export_elements(private_members, lst)

        self._end_class(lst)
        self._export_footer(lst)

    def is_public_element(self):
        return True


class QmlComponent(QmlBaseComponent):
    """A component inside a QmlClass"""

    def __init__(self, name):
        QmlBaseComponent.__init__(self, name)
        self.comment = None

    def _export_content(self, lst):
        component_id = self.get_component_id()
        if component_id:
            if self.comment:
                lst.append(self.comment)

            lst.append("%s %s;" % (self.class_name, component_id))

        # Export child components with the top-level component. This avoids
        # very deep nesting in the generated documentation.
        self._export_elements(self.elements, lst, filter=lambda x:
                              isinstance(x, QmlComponent))

    def get_component_id(self):
        # Returns the id of the component, if it has one
        for attr in self.get_attributes():
            if attr.name == "id":
                return attr.value
        return None

    def is_public_element(self):
        return False


class QmlArgument(object):
    def __init__(self, name):
        self.type = ""
        self.name = name
        self.default_value = None

    def __str__(self):
        if self.type == "":
            return self.name + self.default_value_string()
        else:
            return self.type + " " + self.name + self.default_value_string()

    def default_value_string(self):
        if self.default_value is None:
            return ''
        else:
            return ' = {}'.format(self.default_value)

    def is_public_element(self):
        return True


class QmlAttribute(object):
    def __init__(self):
        self.name = ""
        self.value = ""
        self.type = "var"
        self.doc = ""

    def __str__(self):
        if self.name != "id":
            lst = []
            if len(self.doc) > 0:
                lst.append(self.doc)
            lst.append(self.type + " " + self.name + ";")
            return "\n".join(lst)
        else:
            return ""

    def is_public_element(self):
        return False


class QmlProperty(object):
    type_rx = re.compile(TYPE_RX)

    DEFAULT_PROPERTY_COMMENT = "/** @remark This is the default property */"
    READONLY_PROPERTY_COMMENT = "/** @remark This property is read-only */"

    def __init__(self):
        self.type = ""
        self.is_default = False
        self.is_readonly = False
        self.name = ""
        self.doc = ""
        self.doc_is_inline = False

    def __str__(self):
        self.post_process_doc()
        lst = []
        if not self.doc_is_inline:
            lst.append(self.doc + "\n")
        if self.is_default:
            lst.append(self.DEFAULT_PROPERTY_COMMENT + "\n")
        elif self.is_readonly:
            lst.append(self.READONLY_PROPERTY_COMMENT + "\n")
        lst.append("Q_PROPERTY(%s %s READ dummyGetter_%s_ignore)"
            % (self.type, self.name, self.name))
        if self.doc_is_inline:
            lst.append(" " + self.doc)
        return "".join(lst)

    def post_process_doc(self):
        self.doc, self.type = post_process_type(self.type_rx, self.doc, self.type)

    def is_public_element(self):
        # Doxygen always adds Q_PROPERTY items as public members.
        return True


class QmlFunction(object):
    doc_arg_rx = re.compile(r"[@\\]param" + TYPE_RX + r"\s+(?P<name>\w+)")
    return_rx = re.compile(r"[@\\]returns?" + TYPE_RX)

    def __init__(self):
        self.type = "void"
        self.name = ""
        self.doc = ""
        self.doc_is_inline = False
        self.args = []

    def __str__(self):
        self.post_process_doc()
        arg_string = ", ".join([str(x) for x in self.args])
        lst = []
        if not self.doc_is_inline:
            lst.append(self.doc + "\n")
        lst.append("%s %s(%s);" % (self.type, self.name, arg_string))
        if self.doc_is_inline:
            lst.append(" " + self.doc)
        return "".join(lst)

    def post_process_doc(self):
        def repl(match):
            # For each argument with a specified type, update arg.type and return a typeless @param line
            type = match.group("type")
            name = match.group("name")
            for arg in self.args:
                if arg.name == name:
                    arg.type = type
                    break
            else:
                logging.warning("In function %s(): Unknown argument %s" % (self.name, name))
            return "@param %s" % name

        self.doc = self.doc_arg_rx.sub(repl, self.doc)
        self.doc, self.type = post_process_type(self.return_rx, self.doc, self.type)

    def is_public_element(self):
        return True


class QmlSignal(object):
    def __init__(self):
        self.name = ""
        self.doc = ""
        self.doc_is_inline = False
        self.args = []

    def __str__(self):
        arg_string = ", ".join([str(x) for x in self.args])
        lst = []
        if not self.doc_is_inline:
            lst.append(self.doc + "\n")
        lst.append("Q_SIGNALS: void %s(%s); " % (self.name, arg_string))
        if self.doc_is_inline:
            lst.append(self.doc + "\n")
        # Appending "public:" here makes it possible to declare a signal without
        # turning all functions defined after into signals.
        # It could be replaced with the use of Q_SIGNAL, but my version of
        # Doxygen (1.8.4) does not support it
        lst.append("public:")
        return "".join(lst)

    def is_public_element(self):
        # Doxygen always adds Q_SIGNALS items as public members.
        return True



from collections import namedtuple
import re


COMMENT = "comment"
ICOMMENT = "inline_comment"
STRING = "string"
ELEMENT = "element"
BLOCK_START = "block_start"
BLOCK_END = "block_end"
CHAR = "char"
KEYWORD = "keyword"
IMPORT = "import"
PRAGMA = "pragma"
COMPONENT = "component"
ATTRIBUTE = "attribute"

# not a doxy comment
PLAIN_COMMENT_RX = re.compile("/[/*][^/!*]")


def is_doxy_comment_token(token):
    return token.type == COMMENT and not PLAIN_COMMENT_RX.match(token.value)


class LexerError(Exception):
    def __init__(self, msg, idx):
        Exception.__init__(self, msg)
        self.idx = idx


Token = namedtuple("Token", ["type", "value", "idx"])


class Tokenizer(object):
    def __init__(self, token_type, rx):
        self.token_type = token_type
        self.rx = rx

    def __call__(self, lexer, matched_str):
        lexer.append_token(self.token_type, matched_str)


class Lexer(object):
    def __init__(self, text):
        # Tokens that start at the first non-whitespace character in a line
        self.tokenizers_newline = [
            Tokenizer(COMPONENT, re.compile(r"([-\w\.]+)\s*{")),  # a component
            Tokenizer(ATTRIBUTE, re.compile(r"([-\w\.]+)\s*:")),  # an attribute
            ]

        self.tokenizers = [
            Tokenizer(ICOMMENT, re.compile(r"/\*[!*]<.*?\*/", re.DOTALL)),
            Tokenizer(ICOMMENT, re.compile(r"//[/!]<.*")),
            Tokenizer(COMMENT, re.compile(r"/\*.*?\*/", re.DOTALL)),
            Tokenizer(COMMENT, re.compile(r"//.*")),
            # A double quote, then either:
            # - anything but a double quote or a backslash
            # - an escaped char (\n, \t...)
            # then a double quote
            Tokenizer(STRING, re.compile(r'("([^\\"]|(\\.))*")')),
            Tokenizer(BLOCK_START, re.compile("{")),
            Tokenizer(BLOCK_END, re.compile("}")),
            Tokenizer(IMPORT, re.compile(r"import\s+.*")),
            Tokenizer(PRAGMA, re.compile(r"pragma\s+\w.*")),
            Tokenizer(KEYWORD, re.compile(r"(default\s+property|property|readonly\s+property|signal)\s+")),
            Tokenizer(KEYWORD, re.compile(r"(function)\s+[^(]")),  # a named function
            Tokenizer(ELEMENT, re.compile(r"\w[\w.<>]*")),
            Tokenizer(CHAR, re.compile(".")),
            ]
        self.text = text.replace('\\\n', '\n')
        self.idx = 0
        self.newline = False
        self.tokens = []

    def tokenize(self):
        while True:
            self.advance()
            if self.idx == len(self.text):
                break
            self.apply_tokenizers()
        self.fixup_tokens()

    def advance(self):
        self.newline = False
        if self.idx == 0:
            # Process start-of-file as newline.
            self.newline = True

        while self.idx < len(self.text):
            if self.text[self.idx] == '\n':
                self.newline = True
                self.idx += 1
            elif self.text[self.idx].isspace():
                self.idx += 1
            else:
                break

    def apply_tokenizers(self):
        if self.newline:
            for tokenizer in self.tokenizers_newline:
                match = tokenizer.rx.match(self.text, self.idx)

                if not match:
                    continue

                if len(match.groups()) > 0:
                    tokenizer(self, match.group(1))
                    self.idx = match.end(1)
                    return
                else:
                    tokenizer(self, match.group(0))
                    self.idx = match.end(0)
                    return

        for tokenizer in self.tokenizers:
            match = tokenizer.rx.match(self.text, self.idx)

            if not match:
                continue

            if len(match.groups()) > 0:
                tokenizer(self, match.group(1))
                self.idx = match.end(1)
                return
            else:
                tokenizer(self, match.group(0))
                self.idx = match.end(0)
                return

        raise LexerError("No lexer matched", self.idx)

    def fixup_tokens(self):
        for idx, token in enumerate(self.tokens):
            # Fix tokenization of a property named "property". For example:
            #   property string property: "foo"
            if (token.type == KEYWORD and token.value == "property" and idx > 1 and
                    self.tokens[idx - 1].type == ELEMENT and
                    self.tokens[idx - 2].type == KEYWORD and self.tokens[idx - 2].value.endswith("property")):
                self.tokens[idx] = Token(ELEMENT, token.value, token.idx)
            if token.type == ICOMMENT and idx > 1:
                self.move_inline_comments(idx)

    def move_inline_comments(self, start_idx):
        """
        Move inline comments ahead of their parent KEYWORD. This way they get
        properly handed over to the Qml* object type handlers which can do with
        them as they wish.
        """
        # Iterate backwards looking for a KEYWORD. As a sanity measure we only
        # search back up to 20 tokens or until an "invalid" token is found.
        end_idx = max(start_idx - 20, 0)
        for idx, token in enumerate(self.tokens[start_idx - 1:end_idx:-1]):
            if token.type == KEYWORD:
                ins_idx = start_idx - idx - 1
                if ins_idx <= 0:
                    return
                break
            if token.type in (COMMENT, ICOMMENT, IMPORT, PRAGMA):
                return

        # Final sanity check for a misplaced inline comment
        previous_token = self.tokens[ins_idx - 1]
        if previous_token.type == ICOMMENT or is_doxy_comment_token(previous_token):
            return

        self.tokens.insert(ins_idx, self.tokens.pop(start_idx))

    def append_token(self, type, value):
        self.tokens.append(Token(type, value, self.idx))



#!/usr/bin/env python3
import os
import sys
import unittest


if __name__ == "__main__":
    current_dir = os.path.dirname(__file__)
    args = [sys.argv[0]] + ["discover", "--start-directory", current_dir, "--pattern", "*testcase.py"] + sys.argv[1:]
    unittest.main(module=None, argv=args)
# vi: ts=4 sw=4 et



#!/usr/bin/env python3
import os

from setuptools import setup

from doxyqml import __version__, DESCRIPTION


def read_readme():
    root_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(root_dir, 'README.md'), encoding='utf-8') as f:
        return f.read()


setup(name="doxyqml",
      version=__version__,
      description=DESCRIPTION,
      long_description=read_readme(),
      long_description_content_type="text/markdown",
      author="Aurélien Gâteau",
      author_email="mail@agateau.com",
      license="BSD",
      platforms=["any"],
      url="http://agateau.com/projects/doxyqml",
      packages=["doxyqml"],
      entry_points={
          "console_scripts": [
              "doxyqml = doxyqml.main:main",
          ],
      },
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Environment :: Plugins",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Topic :: Documentation",
      ])
